#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['frappuccino', 'frappuccino.tests']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['frappuccino = frappuccino:main']}

setup(name='frappuccino',
      version='0.0.5',
      description='Frappuccino',
      author='Matthias Bussonnier',
      author_email='bussonniermatthias@gmail.com',
      url='https://github.com/Carreau/frappuccino',
      packages=packages,
      package_data=package_data,
      entry_points=entry_points,
      python_requires='>=3.5',
     )
